﻿/* 数学ライブラリは多分使うと思う */
#include <math.h>

/* int 型の最大値 INT_MAX は limits.h の中で定義されている */
#include <limits.h>

/* (x, y) に色 c で点を打つ関数 (他で定義している) */
extern void point(int, int, const double *);

/* ウィンドウサイズを変更する (他で定義している) */
extern void size(int, int);

/* 背景色 (デフォルトは白) を変更する (他で定義している) */
extern void background(double, double, double);

/* 画面を消去する (他で定義している) */
extern void clear(void);

/* 直前にタイプしたキーの文字を得る (他で定義している) */
extern unsigned char getkey(void);

/* ←:100 ↑:101 →:102 ↓:103 を返す (他で定義している) */
extern int getspecial(void);

/* 経過時間を ms で返す (他で定義している) */
extern int elapsed(void);

/* 描画前の初期設定 */
void setup(void)
{
  /*
  ** ここは最初に draw() を実行する前に一度だけ実行される
  ** 背景色の変更など描画前に行いたい処理があればここに書く
  */
}

/* 線分を描く */
static void line(int x0, int y0, int x1, int y1, const double *c)
{
  int x, y, ix, iy, a, ax, ay, b, bx, by, d, dx, dy, e;
  
  ix = iy = 1;
  
  if ((dx = x1 - (x = x0)) < 0)
  {
    dx = -dx;
    ix = -ix;
  }
  
  if ((dy = y1 - (y = y0)) < 0)
  {
    dy = -dy;
    iy = -iy;
  }
  
  if (dx < dy)
  {
    a = -(dx + dx);
    b = (d = dy) + dy;
    ay = iy;
    bx = ix;
    ax = by = 0;
  }
  else {
    a = -(dy + dy);
    b = (d = dx) + dx;
    ax = ix;
    by = iy;
    ay = bx = 0;
  }    
  
  for (e = d;;)
  {
    point(x, y, c);
    if (--d < 0) break;
    x += ax;
    y += ay;
    if ((e += a) <= 0)
    {
      x += bx;
      y += by;
      e += b;
    }
  }
}

/*
** 座標変換：ベクトル v2 ← ベクトル v1 に行列 m を掛けた結果
*/
static void transform(double v2[], double m[][4], double v1[])
{
  int j;

  for (j = 0; j < 4; ++j)
  {
    v2[j] = m[j][0] * v1[0]
          + m[j][1] * v1[1]
          + m[j][2] * v1[2]
          + m[j][3] * v1[3];
  }
}

/*
** 行列の積：行列 m3 ← 行列 m1 と行列 m2 の積
*/
static void multiply(double m3[][4], double m1[][4], double m2[][4])
{
  int k;

  for (k = 0; k < 4; ++k)
  {
    int j;

    for (j = 0; j < 4; ++j)
    {
      m3[k][j] = m1[k][0] * m2[0][j]
               + m1[k][1] * m2[1][j]
               + m1[k][2] * m2[2][j]
               + m1[k][3] * m2[3][j];
    }
  }
}

/*
** 平行移動：行列 m ← (x, y, z) に平行移動する変換行列
*/
static void setTranslation(double m[][4], double x, double y, double z)
{
  /*
  ** ４×４要素の配列 m の個々の要素に値を設定するプログラムを書く
  */
  m[0][0] = 1.0;  m[0][1] = 0.0;  m[0][2] = 0.0;  m[0][3] = x;
  m[1][0] = 0.0;  m[1][1] = 1.0;  m[1][2] = 0.0;  m[1][3] = y;
  m[2][0] = 0.0;  m[2][1] = 0.0;  m[2][2] = 1.0;  m[2][3] = z;
  m[3][0] = 0.0;  m[3][1] = 0.0;  m[3][2] = 0.0;  m[3][3] = 1.0;
}

/*
** 拡大縮小：行列 m ← 原点中心に (a, b, c) 倍する変換行列
*/
static void setScale(double m[][4], double a, double b, double c)
{
  /*
  ** ４×４要素の配列 m の個々の要素に値を設定するプログラムを書く
  */
  m[0][0] = a;    m[0][1] = 0.0;  m[0][2] = 0.0;  m[0][3] = 0.0;
  m[1][0] = 0.0;  m[1][1] = b;    m[1][2] = 0.0;  m[1][3] = 0.0;
  m[2][0] = 0.0;  m[2][1] = 0.0;  m[2][2] = c;    m[2][3] = 0.0;
  m[3][0] = 0.0;  m[3][1] = 0.0;  m[3][2] = 0.0;  m[3][3] = 1.0;
}

#define PI 3.14159265358979323846                       /* 円周率π */

/*
** x 軸中心の回転：行列 m ← x 軸を中心にして angle 度回転する変換行列
*/
static void setRotationX(double m[][4], double angle)
{
  /*
  ** ４×４要素の配列 m の個々の要素に X 軸中心の回転行列の値を設定する
  */
  double rad = angle * PI / 180.0;
  double c = cos(rad);
  double s = sin(rad);

  m[0][0] = 1.0;  m[0][1] = 0.0;  m[0][2] = 0.0;  m[0][3] = 0.0;
  m[1][0] = 0.0;  m[1][1] = c;    m[1][2] = -s;   m[1][3] = 0.0;
  m[2][0] = 0.0;  m[2][1] = s;    m[2][2] = c;    m[2][3] = 0.0;
  m[3][0] = 0.0;  m[3][1] = 0.0;  m[3][2] = 0.0;  m[3][3] = 1.0;
}

/*
** y 軸中心の回転：行列 m ← y 軸を中心にして angle 度回転する変換行列
*/
static void setRotationY(double m[][4], double angle)
{
  /*
  ** ４×４要素の配列 m の個々の要素に Y 軸中心の回転行列の値を設定する
  */
  double rad = angle * PI / 180.0;
  double c = cos(rad);
  double s = sin(rad);

  m[0][0] = c;    m[0][1] = 0.0;  m[0][2] = s;    m[0][3] = 0.0;
  m[1][0] = 0.0;  m[1][1] = 1.0;  m[1][2] = 0.0;  m[1][3] = 0.0;
  m[2][0] = -s;   m[2][1] = 0.0;  m[2][2] = c;    m[2][3] = 0.0;
  m[3][0] = 0.0;  m[3][1] = 0.0;  m[3][2] = 0.0;  m[3][3] = 1.0;
}

/*
** z 軸中心の回転：行列 m ← z 軸を中心にして angle 度回転する変換行列
*/
static void setRotationZ(double m[][4], double angle)
{
  /*
  ** ４×４要素の配列 m の個々の要素に Z 軸中心の回転行列の値を設定する
  */
  double rad = angle * PI / 180.0;
  double c = cos(rad);
  double s = sin(rad);

  m[0][0] = c;    m[0][1] = -s;   m[0][2] = 0.0;  m[0][3] = 0.0;
  m[1][0] = s;    m[1][1] = c;    m[1][2] = 0.0;  m[1][3] = 0.0;
  m[2][0] = 0.0;  m[2][1] = 0.0;  m[2][2] = 1.0;  m[2][3] = 0.0;
  m[3][0] = 0.0;  m[3][1] = 0.0;  m[3][2] = 0.0;  m[3][3] = 1.0;
}

/*
** 任意軸中心の回転
**     ベクトル (x, y, z) を中心に angle 度回転する
**     変換行列を r に設定する
*/
static void setRotation(double r[][4], double x, double y, double z, double angle)
{
  /*
  ** Step 1. (x, y, z) を方向余弦（単位行列）に変換する
  ** Step 2. angle の単位を radian に変換する
  ** Step 3. ４×４要素の配列 m の個々の要素に方向余弦中心の回転行列の値を設定する
  */
  double a = sqrt(x * x + y * y + z * z);

  if (a != 0.0)
  {
    double l  = x / a, m  = y / a, n  = z / a;
    double l2 = l * l, m2 = m * m, n2 = n * n;
    double lm = l * m, mn = m * n, nl = n * l;

    double rad = angle * PI / 180.0;
    double c = cos(rad), c1 = 1.0 - c;
    double s = sin(rad);

    r[0][0] = (1.0 - l2) * c + l2;
    r[0][1] = lm * c1 - n * s;
    r[0][2] = nl * c1 + m * s;
    r[0][3] = 0.0;

    r[1][0] = lm * c1 + n * s;
    r[1][1] = (1.0 - m2) * c + m2;
    r[1][2] = mn * c1 - l * s;
    r[1][3] = 0.0;

    r[2][0] = nl * c1 - m * s;
    r[2][1] = mn * c1 + l * s;
    r[2][2] = (1.0 - n2) * c + n2;
    r[2][3] = 0.0;

    r[3][0] = 0.0;
    r[3][1] = 0.0;
    r[3][2] = 0.0;
    r[3][3] = 1.0;
  }
}

/*
** 透視変換
**     画角 fovy°で投影を行なう透視変換行列を m に設定する
*/
static void setPerspective(double m[][4], double fovy, double aspect, double zNear, double zFar)
{
  /*
  ** (1) zNear の位置にある前方面と zFar の位置にある後方面の間の空間にある図形を
  **     画角 fovy°で前方面の位置にあるスクリーンに投影する変換行列を m に求める
  **     画面の縦横比は aspect とする
  */
}

/*
** ビュー変換
**     視点 (ex, ey, ez) から目標点 (tx, ty, tz) を上方向を (ux, uy, uz) として
**     見る変換行列を m に設定する
*/
static void lookAt(double m[][4],
  double ex, double ey, double ez,
  double tx, double ty, double tz,
  double ux, double uy, double uz
  )
{
  /*
  ** Step 1. 視点の位置と逆方向に平行移動する変換行列を求める
  ** Step 2. 視点と目標点の位置から求まる視線の回転行列を求める
  ** Step 3. これらを掛けて、結果を m に格納する
  */
  double l;

  // z 軸 = e - t
  tx = ex - tx;
  ty = ey - ty;
  tz = ez - tz;
  l = sqrt(tx * tx + ty * ty + tz * tz);
  if (l == 0.0) return;
  m[2][0] = tx / l;
  m[2][1] = ty / l;
  m[2][2] = tz / l;

  // x 軸 = u x z 軸
  tx = uy * m[2][2] - uz * m[2][1];
  ty = uz * m[2][0] - ux * m[2][2];
  tz = ux * m[2][1] - uy * m[2][0];
  l = sqrt(tx * tx + ty * ty + tz * tz);
  if (l == 0.0) return;
  m[0][0] = tx / l;
  m[0][1] = ty / l;
  m[0][2] = tz / l;

  // y 軸 = z 軸 x x 軸
  m[1][0] = m[2][1] * m[0][2] - m[2][2] * m[0][1];
  m[1][1] = m[2][2] * m[0][0] - m[2][0] * m[0][2];
  m[1][2] = m[2][0] * m[0][1] - m[2][1] * m[0][0];

  // 平行移動
  m[0][3] = -(ex * m[0][0] + ey * m[0][1] + ez * m[0][2]);
  m[1][3] = -(ex * m[1][0] + ey * m[1][1] + ez * m[1][2]);
  m[2][3] = -(ex * m[2][0] + ey * m[2][1] + ez * m[2][2]);

  // 残り
  m[3][0] = m[3][1] = m[3][2] = 0.0;
  m[3][3] = 1.0;
}

/*
** 投影によって画面上の座標値を求める
**     v: 座標値の配列（頂点表）, m: 変換行列
**     s: 画面上の座標値の配列, n: 座標数（頂点数）
*/
static void projection(double t[][4], double m[][4], double v[][4], int n)
{
  /*
  ** 引数 v で与えられた n 個の４要素の配列（２次元配列）の
  ** 個々の要素を引数 m で与えられた４×４要素の配列（変換行列）
  ** で座標変換し，引数 t で与えられた配列の各要素に格納する
  */
  int i;

  for (i = 0; i < n; ++i) transform(t[i], m, v[i]);
}

/*
** ワイヤーフレームの描画（稜線表の稜線を v に与えられた座標値で描画）
**     vp: ビューポート変換行列
**     s: クリッピング座標値, e: 稜線表, ne: 稜線数
*/
static void wireframe(double vp[4][4], double s[][4], int e[][2], int ne)
{
  static const double black[] = { 0.0, 0.0, 0.0 };
  int i;

  for (i = 0; i < ne; ++i)
  {
  	/* 線分の端点の番号を取り出す */
    int v0 = e[i][0], v1 = e[i][1];
    /* 端点の実座標を求めて整数化する */
    int x0 = (int)(s[v0][0] * vp[0][0] / s[v0][3] + vp[0][3] + 0.5);
    int y0 = (int)(s[v0][1] * vp[1][1] / s[v0][3] + vp[1][3] + 0.5);
    int x1 = (int)(s[v1][0] * vp[0][0] / s[v1][3] + vp[0][3] + 0.5);
    int y1 = (int)(s[v1][1] * vp[1][1] / s[v1][3] + vp[1][3] + 0.5);
    
    line(x0, y0, x1, y1, black);
  }
}

/* 頂点の座標値（頂点表のデータ） */
static double vertex[][4] =
{
  /*
  ** (2) 頂点表のデータ (同次座標) をここに書く
  */
  { -4.0,  4.0,  2.0,  1.0 }, /* 頂点 a */
  {  4.0,  4.0,  2.0,  1.0 }, /* 頂点 b */
  {  3.0,  3.0,  0.0,  1.0 }, /* 頂点 c */
  {  4.0,  4.0, -2.0,  1.0 }, /* 頂点 d */
  { -4.0,  4.0, -2.0,  1.0 }, /* 頂点 e */
  { -3.0,  3.0,  0.0,  1.0 }, /* 頂点 f */
  {  0.0,  0.0,  0.0,  1.0 }, /* 頂点 g */
};

/* 線分の端点番号（稜線表のデータ） */
static int edge[][2] =
{
  /*
  ** (3) 稜線表のデータをここに書く
  */
  { 2, 3 }, /* 稜線ア (c-d) */
  { 3, 4 }, /* 稜線イ (d-e) */
  { 4, 5 }, /* 稜線ウ (e-f) */
  { 5, 2 }, /* 稜線エ (f-c) */
  { 2, 1 }, /* 稜線オ (c-b) */
  { 1, 0 }, /* 稜線カ (b-a) */
  { 0, 5 }, /* 稜線キ (a-f) */
  { 0, 6 }, /* 稜線ク (a-g) */
  { 1, 6 }, /* 稜線ケ (b-g) */
  { 3, 6 }, /* 稜線コ (d-g) */
  { 4, 6 }, /* 稜線サ (e-g) */
};

#define NVERTEX (sizeof vertex / sizeof vertex[0])      /* 頂点の数 */
#define NEDGE (sizeof edge / sizeof edge[0])            /* 稜線の数 */

/* 視点座標 */
static double viewing[NVERTEX][4];

/* クリッピング座標 */
static double clipping[NVERTEX][4];

/*
** 図形の描画
*/
void draw(int width, int height)
{
  double rotation[4][4];                    /* 物体の回転の変換行列 */
  double spin[4][4];                        /* スピン用の変換行列　 */
  double model[4][4];                       /* モデル変換行列　　　 */
  double view[4][4];                        /* ビュー変換行列　　　 */
  double modelview[4][4];                   /* モデルビュー変換行列 */
  double perspective[4][4];                 /* 透視投影変換行列　　 */
  double scale[4][4];                       /* 拡大縮小の変換行列　 */
  double translation[4][4];                 /* 平行移動の変換行列　 */
  double viewport[4][4];                    /* ビューポート変換行列 */

  /* 図形を回転させるための回転角 */
  const double r = (elapsed() % 6000) * 0.06;

  /*
  ** ビューポート変換行列を求める
  */

  /* scale にウィンドウサイズにあわせて図形を拡大縮小する変換行列を求める */
  setScale(scale, (double)width * 0.5, (double)height * 0.5, (double)INT_MAX * 0.5);
  /* translation に図形がウィンドウに収まるように平行移動する変換行列を求める */
  setTranslation(translation, (double)width * 0.5, (double)height * 0.5, (double)INT_MAX * 0.5);
  /* 拡大縮小の変換行列に平行移動の変換行列を掛けてビューポート変換行列を求める */
  multiply(viewport, translation, scale);

  /*
  ** (4) 必要に応じて以下を書き換える
  */

  /*
  ** ビュー変換行列を求める
  */
  lookAt(view, 0.0, 15.0, 20.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);

  /*
  ** 透視投影変換行列を求める
  */
  setPerspective(perspective, 30.0, (double)width / (double)height, 5.0, 50.0);

  /*
  ** 回転アニメーション用の変換行列を求める
  */
  setRotationY(spin, r);

  /*
  ** 一つ目の図形のモデル変換行列を求める
  */
  setRotationZ(rotation, 45.0);             /* Z 軸中心に 45°回転  */

  /*
  ** 一つ目の図形の座標変換と描画
  */
  multiply(model, spin, rotation);                    /* モデル変換 */
  multiply(modelview, view, model);                   /* ビュー変換 */
  projection(viewing, modelview, vertex, NVERTEX);      /* 視点座標 */
  projection(clipping, perspective, viewing, NVERTEX);      /* 投影 */
  wireframe(viewport, clipping, edge, NEDGE);               /* 描画 */

  /*
  ** 二つ目の図形のモデル変換行列を求める
  */
  setRotationZ(rotation, -135.0);         /* Z 軸中心に -135°回転  */

  /*
  ** 二つ目の図形の座標変換と描画
  */
  multiply(model, spin, rotation);                    /* モデル変換 */
  multiply(modelview, view, model);                   /* ビュー変換 */
  projection(viewing, modelview, vertex, NVERTEX);      /* 視点座標 */
  projection(clipping, perspective, viewing, NVERTEX);      /* 投影 */
  wireframe(viewport, clipping, edge, NEDGE);               /* 描画 */
}
